Util   = require 'util'
Path   = require 'path'
Fs     = require 'fs'
Crypto = require 'crypto'
Mime   = require('simple-mime') 'application/octect-stream'

methods = ['GET', 'HEAD']

# Public: Sets up a Stack handler for serving static files.
#
# root    - String path to the root directory to serve static files from.
# options - Optional Hash of options:
#           default - String that specifies the default filename for a 
#                     directory.
#
# Returns a Stack handler function.
module.exports = (root, options = {}) ->
  root        = Path.normalize root
  defaultFile = options.default
  delete options.default
  if !options.etag? then options.etag = module.exports.etag
  (req, res, next) ->
    if methods.indexOf(req.method) == -1
      res.writeHead 405, accept: 'GET, HEAD'
      return res.end()
    if defaultFile && req.url.match(/\/$/)
      req.url = "#{req.url}#{defaultFile}"

    options.file = "#{root}#{req.url}"
    module.exports.serve req, res, options, (err) -> 
      next() if err # pass on the request if it's a 404

# Public: Serves a given file from the file system.
#
# req     - The http.ServerRequest instance.
# res     - The http.ServerResponse instance.
# options - Hash with the parameters defining the file to serve.
#           file - String path to the file on the file system.
#           etag - Optional Function for generating an Etag.  See writeHead()
# next    - An optional Function callback to be called when the file has 
#           finished streaming.  If the file doesn't exist, an err argument
#           from fs.stat() is passed.
#
# Returns nothing.
module.exports.serve = (req, res, options = {}, next) ->
  file = options.file
  delete options.file
  Fs.stat file, (err, stats) ->
    if err
      next err if next
      return
    module.exports.writeHead req, res, file, stats, options, (status) ->
      if status < 200 or status > 299 or req.method == 'HEAD'
        res.end()
        next() if next
      else
        Util.pump Fs.createReadStream(file), res, next or empty_cb
empty_cb = ->

# Public: Sets up the headers for the HTTP response for the given served file.
#
# res     - The http.ServerResponse instance.
# req     - The http.ServerRequest instance.
# file    - String path to the file on the file system.
# stats   - fs.Stats instance of the file.
# options - Options Hash:
#           status - The default Integer HTTP status code.
#           etag   - A function for generating an ETag, or 'true' to use the
#                    default ETag generator.  If an ETag is generated, it is
#                    checked against an 'if-none-match' request header.
#
# Returns a Hash of headers.
module.exports.writeHead = (req, res, file, stats, options = {}, cb) ->
  headers = module.exports.defaultHeaders file, stats
  status  = options.status or 200
  ifmod   = req.headers['if-modified-since']

  if ifmod
    since = Date.parse ifmod
    if since and since > stats.mtime.getTime()
      return not_modified res, headers, cb

  ifunmod = req.headers['if-unmodified-since']
  if ifunmod
    since = Date.parse ifunmod
    if since and since < stats.mtime.getTime()
      headers['content-length'] = 0
      res.writeHead 412, headers
      return cb(412)

  if options.etag
    options.etag file, stats, (etag) ->
      headers.etag = etag
      if etag_matches req.headers['if-none-match'], etag
        return not_modified res, headers, cb
      res.writeHead status, headers
      cb(status)
  else
    res.writeHead status, headers
    cb(status)

# Public: Sets up the default headers for the HTTP response for the given 
# served file.
#
# file  - String path to the file on the file system.
# stats - fs.Stats instance of the file.
#
# Returns a Hash of headers.
module.exports.defaultHeaders = (file, stats) ->
  'accept-ranges':  'bytes'
  'content-type':   Mime(file)
  'content-length': stats.size
  'last-modified':  stats.mtime.toUTCString()

# Generates an ETag of the given file by creating a SHA1 hash from the File 
# inode + size + mtime.  This should probably not be used if you're serving 
# files from multiple servers.
#
# file  - String path to the file on the file system.
# stats - fs.Stats instance of the file.
# next  - An optional Function callback with the String ETag.
#
# Returns nothing.
module.exports.etag = (file, stats, next) ->
  hash = Crypto.createHash 'sha1'
  hash.update stats.ino
  hash.update stats.size
  hash.update stats.mtime
  next "\"#{hash.digest('hex')}\""

# Helper function to write a "304 Not Modified" HTTP response.
#
# res     - The http.ServerResponse instance.
# headers - Hash of HTTP headers.
# cb      - The function callback to be called with the HTTP status code (304).
#
# Returns nothing.
not_modified = (res, headers, cb) ->
  headers['content-length'] = 0
  res.writeHead 304, headers
  cb(304)

# Checks if the requested file's etag matches the requested ETag.
#
# request_etag - String ETag sent in the If-None-Match HTTP request header.
# file_etag    - String ETag of the requested file.
#
# Returns true if the etags match, or false.
etag_matches = (request_etag, file_etag) ->
  request_etag? and file_etag? and request_etag == file_etag